package eastsun.jgvm.module;

import eastsun.jgvm.module.event.Area;
import eastsun.jgvm.module.ram.Accessable;
import eastsun.jgvm.module.ram.Ram;
import eastsun.jgvm.module.ram.RelativeRam;
import eastsun.jgvm.module.ram.RuntimeRam;
import eastsun.jgvm.module.ram.Stack;
import eastsun.jgvm.module.ram.StringRam;

/**
 * ڴģ,ĬάʱڴַѲṩջ֧<p>
 * ͨinstallֶ֧ԴıĿѡ֧<p>
 * GVMͨRamManagerṩڴдRam,Ҫ֪ľṹ<p>
 * @version 2.0 2008/3/19
 * @author Eastsun
 */
public final class RamManager implements Accessable {

    public static final int SIZE_OF_ADDR = 3;
    public static final int SIZE_OF_CHAR = 1;
    public static final int SIZE_OF_INT = 2;
    public static final int SIZE_OF_LONG = 4;
    /**
     * ʱڴĿʼַ
     */
    public static final int START_ADDR = 0x2000;
    private RelativeRam textRam,  graphRam,  bufferRam;
    private ScreenModel screen;
    private RuntimeRam runRam;
    private StringRam strRam;
    private Stack stack;
    private Ram[] rams = new Ram[3];
    private int ramCount;

    public RamManager(RuntimeRam runRam, StringRam strRam, Stack stack) {
        if (runRam == null || strRam == null || stack == null) {
            throw new IllegalArgumentException("param can't be null");
        }
        this.stack = stack;
        install(runRam);
        install(strRam);
    }

    /**
     * ͬgetByte(addr)
     * @param addr ַ
     * @return һ޷charֵʾһbyteֵ
     * @see #getByte(int)
     */
    public char getChar(int addr) {
        return (char) (getByte(addr) & 0xff);
    }

    /**
     * ַָȡֽһlavaеint
     * @param addr ַ
     * @return int
     */
    public short getInt(int addr) {
        return (short) getBytes(addr, SIZE_OF_INT);
    }

    /**
     * ַָȡֽһlavaеļָ
     * @param addr ַ
     * @return ļָ
     */
    public int getAddr(int addr) {
        return getBytes(addr, SIZE_OF_ADDR);
    }

    /**
     * ַָȡֽһlavaеlong
     * @param addr ַ
     * @return long
     */
    public int getLong(int addr) {
        return getBytes(addr, SIZE_OF_LONG);
    }

    /**
     * һlavaеchar,ͬsetByte((byte)c)
     * @param addr ַ
     * @param c charֵ,ڰλЧ
     * @see #setByte(int,byte)
     */
    public void setChar(int addr, char c) {
        setByte(addr, (byte) c);
    }

    /**
     * һlavaеint
     */
    public void setInt(int addr, short i) {
        setBytes(addr, SIZE_OF_INT, i);
    }

    /**
     * һlavaеļָ
     */
    public void setAddr(int addr, int a) {
        setBytes(addr, SIZE_OF_ADDR, a);
    }

    /**
     * һlavaеlong
     */
    public void setLong(int addr, int l) {
        setBytes(addr, SIZE_OF_LONG, l);
    }

    /**
     * ָڴȡcountֽƴһ
     * @param addr ʼַ
     * @param count ֽ
     * @return ֵ
     */
    public int getBytes(int addr, int count) {
        int data = 0;
        while (--count >= 0) {
            data <<= 8;
            data |= (getByte(addr + count) & 0xff);
        }
        return data;
    }

    /**
     * һֵڴcountֵֽ<p>
     * ע:Щڴַп漰ԴĻ,÷֪ͨӦ<p>
     *      ӦԼintersectWithGraph()жϲӦӦ
     * @see #intersectWithGraph(int,int)
     */
    public void setBytes(int addr, int count, int data) {
        while (--count >= 0) {
            setByte(addr++, (byte) data);
            data >>>= 8;
        }
    }

    /**
     * õstringRam
     */
    public StringRam getStringRam() {
        return strRam;
    }

    /**
     * õJLVMʹõStack
     * @return stack
     */
    public Stack getStack() {
        return stack;
    }

    /**
     * õJLVMʹõRuntimeRam
     * @return runtimeRam
     */
    public RuntimeRam getRuntimeRam() {
        return runRam;
    }

    /**
     * õstartʼendڴַĻϵʾ<p>
     * @param start ʼַ,
     * @param end ַ,
     * @return ڴĻʾСǾ;ִڴ鲻Դཻ,򷵻һյArea
     */
    public Area intersectWithGraph(int start, int end) {
        if (graphRam == null) {
            return Area.EMPTY_AREA;
        }
        if (start >= graphRam.getStartAddr() + graphRam.size() || end <= graphRam.getStartAddr()) {
            return Area.EMPTY_AREA;
        }
        if (start < graphRam.getStartAddr()) {
            start = graphRam.getStartAddr();
        }
        if (end > graphRam.getStartAddr() + graphRam.size()) {
            end = graphRam.getStartAddr() + graphRam.size();
        }
        start <<= 3;
        end = (end << 3) - 1;
        int y1 = start / screen.getWidth();
        int x1 = start % screen.getWidth();
        int y2 = end / screen.getWidth();
        int x2 = end % screen.getWidth();
        if (y1 == y2) {
            return new Area(x1, y1, x2 - x1 + 1, 1);
        }
        else {
            return new Area(0, y1, screen.getWidth(), y2 - y1 + 1);
        }
    }

    /**
     * ȡָڴַһֽ,byte
     * @param addr ַ
     * @return byte
     */
    public byte getByte(int addr) {
        //Notice: ʵresetRamAddressʵַʽй
        if (addr >= runRam.getStartAddr()) {
            return runRam.getByte(addr);
        }
        if (addr >= strRam.getStartAddr()) {
            return strRam.getByte(addr);
        }
        for (int index = ramCount - 1; index >= 0; index--) {
            Ram ram = rams[index];
            if (addr >= ram.getStartAddr()) {
                return ram.getByte(addr);
            }
        }
        throw new IndexOutOfBoundsException("ڴԽ:" + addr);
    }

    /**
     * ַΪaddrΪb
     * ע:Щڴַп漰ԴĻ,÷֪ͨӦ<p>
     *      ӦԼintersectWithGraph()жϲӦӦ
     * @param addr ַ
     * @param b 
     * @throws IndexOutOfBoundsException ڴдԽ
     */
    public void setByte(int addr, byte b) {
        //Notice: ʵresetRamAddressʵַʽй
        if (addr >= runRam.getStartAddr()) {
            runRam.setByte(addr, b);
            return;
        }
        if (addr >= strRam.getStartAddr()) {
            strRam.setByte(addr, b);
            return;
        }
        for (int index = ramCount - 1; index >= 0; index--) {
            Ram ram = rams[index];
            if (addr >= ram.getStartAddr()) {
                ram.setByte(addr, b);
                return;
            }
        }
        throw new IndexOutOfBoundsException("ڴдԽ:" + addr);
    }

    /**
     * ڴģ
     */
    public void clear() {
        runRam.clear();
        strRam.clear();
        stack.clear();
        for (int index = 0; index < ramCount; index++) {
            rams[index].clear();
        }

    }

    /**
     * ڴģ,ÿ͵ڴֻܰװһ.<p>
     * ramΪnull׳쳣,ҲıRamManagerκ״̬
     * @param ram Ҫװڴ
     * @throws IllegalStateException Ѿװ͵ڴ
     * @see #uninstall
     */
    public void install(Ram ram) {
        if (ram == null) {
            return;
        }
        switch (ram.getRamType()) {
            case Ram.RAM_RUNTIME_TYPE:
                if (runRam != null) {
                    throw new IllegalStateException("Runtime Ram was installed!");
                }
                runRam = (RuntimeRam) ram;
                break;

            case Ram.RAM_GRAPH_TYPE:
                if (graphRam != null) {
                    throw new IllegalStateException("Graph Ram was installed!");
                }
                graphRam = (RelativeRam) ram;
                screen = graphRam.getScreenModel();
                break;

            case Ram.RAM_BUFFER_TYPE:
                if (bufferRam != null) {
                    throw new IllegalStateException("Buffer Ram was installed!");
                }
                bufferRam = (RelativeRam) ram;
                break;

            case Ram.RAM_STRING_TYPE:
                if (strRam != null) {
                    throw new IllegalStateException("String Ram was installed!");
                }
                strRam = (StringRam) ram;
                break;

            case Ram.RAM_TEXT_TYPE:
                if (textRam != null) {
                    throw new IllegalStateException("Text Ram was installed!");
                }
                textRam = (RelativeRam) ram;
                break;

        }
        resetRamAddress();

    }

    /**
     * жram,ûаװram,ʲôҲᷢ
     * @param type Ҫжصڴ
     */
    public void uninstall(int type) {
        Ram ram = null;
        switch (type) {
            case Ram.RAM_RUNTIME_TYPE:
                ram = runRam;
                runRam = null;
                break;

            case Ram.RAM_GRAPH_TYPE:
                ram = graphRam;
                graphRam = null;
                break;

            case Ram.RAM_BUFFER_TYPE:
                ram = bufferRam;
                bufferRam = null;
                break;

            case Ram.RAM_STRING_TYPE:
                ram = strRam;
                strRam = null;
                break;

            case Ram.RAM_TEXT_TYPE:
                ram = textRam;
                textRam = null;
                break;

        }
        if (ram != null) {
            resetRamAddress();
        }

    }

    private void resetRamAddress() {

        // ڴģ鲼: 0x2000ʼΪRuntimeRam
        //                      0x0000ʼΪgraphRam,bufferRam,textRam,stringRam
        // ӦðʹøʴС,ramsеĴӦðʼַС
        // Notice: ޸,ͬʱҪ޸setBytegetByte

        //ֹЧ
        for (int index = 0; index < rams.length; index++) {
            rams[index] = null;
        }

        ramCount = 0;
        if (runRam != null) {
            //ӵrams
            runRam.setStartAddr(START_ADDR);
        }

        int startAddr = 0;
        if (graphRam != null) {
            graphRam.setStartAddr(startAddr);
            startAddr += graphRam.size();
            rams[ramCount++] = graphRam;
        }

        if (bufferRam != null) {
            bufferRam.setStartAddr(startAddr);
            startAddr += bufferRam.size();
            rams[ramCount++] = bufferRam;
        }

        if (textRam != null) {
            textRam.setStartAddr(startAddr);
            startAddr += textRam.size();
            rams[ramCount++] = textRam;
        }

        if (strRam != null) {
            strRam.setStartAddr(startAddr);
            startAddr += strRam.size();
        //strRamruntimeRamrams֮
        //rams[ramCount++] = strRam;
        }

        if (startAddr > START_ADDR) {
            throw new IllegalStateException(",ڴģô!");
        }
    }
}
